"use strict";

exports.__esModule = true;
exports.Parser = Parser;

var _identifier = require("./identifier");

var _tokentype = require("./tokentype");

var _whitespace = require("./whitespace");

function Parser(options, input, startPos) {
  this.options = options;
  this.isKeyword = _identifier.keywords[6];
  this.isReservedWord = _identifier.reservedWords[6];
  this.input = input;
  this.loadPlugins(this.options.plugins);

  // Set up token state

  // The current position of the tokenizer in the input.
  if (startPos) {
    this.pos = startPos;
    this.lineStart = Math.max(0, this.input.lastIndexOf("\n", startPos));
    this.curLine = this.input.slice(0, this.lineStart).split(_whitespace.lineBreak).length;
  } else {
    this.pos = this.lineStart = 0;
    this.curLine = 1;
  }

  // Properties of the current token:
  // Its type
  this.type = _tokentype.types.eof;
  // For tokens that include more information than their type, the value
  this.value = null;
  // Its start and end offset
  this.start = this.end = this.pos;
  // And, if locations are used, the {line, column} object
  // corresponding to those offsets
  this.startLoc = this.endLoc = this.curPosition();

  // Position information for the previous token
  this.lastTokEndLoc = this.lastTokStartLoc = null;
  this.lastTokStart = this.lastTokEnd = this.pos;

  // The context stack is used to superficially track syntactic
  // context to predict whether a regular expression is allowed in a
  // given position.
  this.context = this.initialContext();
  this.exprAllowed = true;

  // Figure out if it's a module code.
  this.inModule = this.options.sourceType === "module";
  this.strict = this.options.strictMode === false ? false : this.inModule;

  // Used to signify the start of a potential arrow function
  this.potentialArrowAt = -1;

  // Flags to track whether we are in a function, a generator.
  this.inFunction = this.inGenerator = false;
  // Labels in scope.
  this.labels = [];

  // Leading decorators.
  this.decorators = [];

  // Token store.
  this.tokens = [];

  // Comment store.
  this.comments = [];

  // Comment attachment store
  this.trailingComments = [];
  this.leadingComments = [];
  this.bottomRightStack = [];

  // If enabled, skip leading hashbang line.
  if (this.pos === 0 && this.input[0] === "#" && this.input[1] === "!") {
    this.skipLineComment(2);
  }
}

Parser.prototype.extend = function (name, f) {
  this[name] = f(this[name]);
};

// Registered plugins

var plugins = {};

exports.plugins = plugins;
Parser.prototype.loadPlugins = function (plugins) {
  for (var _name in plugins) {
    var plugin = exports.plugins[_name];
    if (!plugin) throw new Error("Plugin '" + _name + "' not found");
    plugin(this, plugins[_name]);
  }
};

Parser.prototype.parse = function () {
  var file = this.startNode();
  var program = this.startNode();
  this.nextToken();
  return this.parseTopLevel(file, program);
};